package org.processmining.starstar.parser;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.io.IOUtils;
import org.processmining.starstar.cep.EsperConditions;
import org.processmining.starstar.elements.XOCActivity;
import org.processmining.starstar.elements.XOCEvent;
import org.processmining.starstar.elements.XOCObject;
import org.processmining.starstar.elements.XOCObjectClass;
import org.processmining.starstar.elements.XOCObjectVersion;
import org.processmining.starstar.elements.XOCRelation;
import org.processmining.starstar.model.EventEdge;
import org.processmining.starstar.model.ModelEdge;
import org.processmining.starstar.model.Subgraph;
import org.processmining.starstar.model.SubgraphComparator;
import org.processmining.starstar.utilities.ModelEdgeComparator;
import org.processmining.starstar.utilities.XOCEventComparator;

public class ParsedXOC {
	public List<String> wantedActivities;
	public List<String> unfilteredWantedActivities;
	public ParsedXOC unfilteredModel;
	public String xocText;
	public Set<String> listOfIdsToKeep;
	public Map<String, XOCActivity> xocActivities;
	public List<XOCEvent> orderedXocEvents;
	public Map<String, XOCEvent> xocEvents;
	public Map<String, XOCObject> xocObjects;
	public Map<String, XOCObjectClass> xocObjectClasses;
	public List<XOCObjectVersion> objectVersions;
	public List<XOCRelation> xocRelations;
	public Map<XOCActivity, Map<XOCActivity, Map<XOCObjectClass, List<EventEdge>>>> eventEdges;
	public Map<XOCObject, List<XOCObject>> objectsIntersections;
	public EsperConditions esperConditions;
	
	public ParsedXOC() {
		wantedActivities = null;
		unfilteredWantedActivities = null;
		listOfIdsToKeep = null;
		initializeModel();
	}
	
	public void cloneIntoUnfiltered(boolean forced) {
		if (unfilteredModel == null || forced) {
			unfilteredModel = new ParsedXOC();
			unfilteredModel.wantedActivities = wantedActivities;
			unfilteredModel.listOfIdsToKeep = listOfIdsToKeep;
			unfilteredModel.xocText = xocText;
			unfilteredModel.xocActivities = xocActivities;
			unfilteredModel.orderedXocEvents = orderedXocEvents;
			unfilteredModel.xocEvents = xocEvents;
			unfilteredModel.xocObjects = xocObjects;
			unfilteredModel.xocObjectClasses = xocObjectClasses;
			unfilteredModel.objectVersions = objectVersions;
			unfilteredModel.xocRelations = xocRelations;
			unfilteredModel.eventEdges = eventEdges;
			unfilteredModel.objectsIntersections = objectsIntersections;
			unfilteredModel.esperConditions = esperConditions;
		}
		if (forced) {
			this.unfilteredWantedActivities = this.wantedActivities;
		}
	}
	
	public void restoreUnfiltered() {
		if (this.wantedActivities != this.unfilteredWantedActivities) {
			this.wantedActivities = this.unfilteredWantedActivities;
			parseSimple(xocText, this.listOfIdsToKeep);
		}
		
		if (unfilteredModel != null) {
			this.listOfIdsToKeep = unfilteredModel.listOfIdsToKeep;
			this.xocText = unfilteredModel.xocText;
			this.xocActivities = unfilteredModel.xocActivities;
			this.orderedXocEvents = unfilteredModel.orderedXocEvents;
			this.xocEvents = unfilteredModel.xocEvents;
			this.xocObjects = unfilteredModel.xocObjects;
			this.xocObjectClasses = unfilteredModel.xocObjectClasses;
			this.objectVersions = unfilteredModel.objectVersions;
			this.xocRelations = unfilteredModel.xocRelations;
			this.eventEdges = unfilteredModel.eventEdges;
			this.objectsIntersections = unfilteredModel.objectsIntersections;
			this.esperConditions = unfilteredModel.esperConditions;
		}
	}
	
	public void initializeModel() {
		xocText = new String();
		xocActivities = new HashMap<String, XOCActivity>();
		orderedXocEvents = new ArrayList<XOCEvent>();
		xocEvents = new HashMap<String, XOCEvent>();
		xocObjects = new HashMap<String, XOCObject>();
		xocObjectClasses = new HashMap<String, XOCObjectClass>();
		objectVersions = new ArrayList<XOCObjectVersion>();
		xocRelations = new ArrayList<XOCRelation>();
		eventEdges = new HashMap<XOCActivity, Map<XOCActivity, Map<XOCObjectClass, List<EventEdge>>>>();
		objectsIntersections = new HashMap<XOCObject, List<XOCObject>>();
		esperConditions = new EsperConditions();
	}
	
	public void setWantedActivities(List<String> wantedActivities) {
		this.wantedActivities = wantedActivities;
		parseSimple(xocText, this.listOfIdsToKeep);
	}
	
	public Map<String, XOCObjectClass> getXocObjectClasses() {
		return this.xocObjectClasses;
	}
	
	public Map<XOCObject, List<XOCObject>> getIntersectedObjects() {
		return this.objectsIntersections;
	}
	
	public void findEventEdges() {
		for (XOCObject object : xocObjects.values()) {
			XOCObjectClass objectClass = object.getObjectClass();
			if (objectClass != null) {
				List<XOCEvent> subscribedEvents = object.getSubscribedEvents();
				Collections.sort(subscribedEvents, new XOCEventComparator());
				//subscribedEvents.sort
				
				int i = 0;
				while (i < subscribedEvents.size()-1) {
					XOCEvent sourceEvent = subscribedEvents.get(i);
					XOCActivity activitySourceEvent = sourceEvent.getActivity();
					if (!eventEdges.containsKey(activitySourceEvent)) {
						eventEdges.put(activitySourceEvent, new HashMap<XOCActivity, Map<XOCObjectClass, List<EventEdge>>>());
					}
					int j = i+1;
					XOCEvent targetEvent = subscribedEvents.get(j);
					XOCActivity activityTargetEvent = targetEvent.getActivity();
					if (!eventEdges.get(activitySourceEvent).containsKey(activityTargetEvent)) {
						eventEdges.get(activitySourceEvent).put(activityTargetEvent, new HashMap<XOCObjectClass, List<EventEdge>>());
					}
					if (!eventEdges.get(activitySourceEvent).get(activityTargetEvent).containsKey(objectClass)) {
						eventEdges.get(activitySourceEvent).get(activityTargetEvent).put(objectClass, new ArrayList<EventEdge>());
					}
					EventEdge eventEdge = new EventEdge(sourceEvent, targetEvent, object);
					eventEdge.setWeight(object.getWeight());
					eventEdge.setPerformance((targetEvent.getTimestamp().getTime()-sourceEvent.getTimestamp().getTime())/1000.0);
					//eventEdge.printResume();
					eventEdges.get(activitySourceEvent).get(activityTargetEvent).get(objectClass).add(eventEdge);

					i++;
				}
			}
		}
	}
	
	public ModelEdge eventEdgesAggregation(XOCActivity sourceActivity, XOCActivity targetActivity, XOCObjectClass objectClass, List<EventEdge> eventEdges) {
		int edgesCount = 0;
		double edgesSumWeights = 0.0;
		double edgesAggregationPerformance = 0.0;
		
		for (EventEdge e : eventEdges) {
			edgesCount++;
			edgesSumWeights += e.getWeight();
			edgesAggregationPerformance += e.getPerformance();
		}
		
		if (edgesCount > 0) {
			edgesAggregationPerformance = edgesAggregationPerformance / edgesCount;
		}
		
		ModelEdge modelEdge = new ModelEdge(sourceActivity, targetActivity, objectClass, eventEdges);
		modelEdge.setCount(edgesCount);
		modelEdge.setWeight(edgesSumWeights);
		modelEdge.setPerformanceAggregation(edgesAggregationPerformance);
		//modelEdge.printResume();
		
		sourceActivity.addOutputModelEdge(modelEdge);
		targetActivity.addInputModelEdge(modelEdge);
		
		return modelEdge;
	}
	
	public void mineModel() {
		for (XOCActivity sourceActivity : eventEdges.keySet()) {
			for (XOCActivity targetActivity : eventEdges.get(sourceActivity).keySet()) {
				for (XOCObjectClass objectClass : eventEdges.get(sourceActivity).get(targetActivity).keySet()) {
					List<EventEdge> eventEdge = eventEdges.get(sourceActivity).get(targetActivity).get(objectClass);
					eventEdgesAggregation(sourceActivity, targetActivity, objectClass, eventEdge);
				}
			}
		}
	}
	
	public void checkEdgesImportance() {
		for (XOCActivity sourceActivity : eventEdges.keySet()) {
			List<ModelEdge> modelEdges = new ArrayList<ModelEdge>(sourceActivity.getOutputModelEdges());
			Collections.sort(modelEdges, new ModelEdgeComparator());
			if (modelEdges.size() > 0) {
				modelEdges.get(0).setStrongestEdgeExitingAnActivity();
			}
		}
		for (XOCActivity targetActivity : eventEdges.keySet()) {
			List<ModelEdge> modelEdges = new ArrayList<ModelEdge>(targetActivity.getInputModelEdges());
			Collections.sort(modelEdges, new ModelEdgeComparator());
			if (modelEdges.size() > 0) {
				modelEdges.get(0).setStrongestEdgeEnteringAnActivity();
			}
		}
	}
	
	public void reviseEdgesWeight() {
		for (XOCActivity sourceActivity : eventEdges.keySet()) {
			List<ModelEdge> modelEdges = new ArrayList<ModelEdge>(sourceActivity.getOutputModelEdges());
			for (ModelEdge edge : modelEdges) {
				edge.reviseWeight();
			}
		}
	}
	
	public void reviseEdgesWeight2() {
		List<Double> valuesBelowHalf = new ArrayList<Double>();
		for (XOCActivity sourceActivity : eventEdges.keySet()) {
			List<ModelEdge> modelEdges = new ArrayList<ModelEdge>(sourceActivity.getOutputModelEdges());
			for (ModelEdge edge : modelEdges) {
				if (edge.getWeight() < 0.5) {
					valuesBelowHalf.add(edge.getWeight());
				}
			}
		}
		Collections.sort(valuesBelowHalf);
		for (XOCActivity sourceActivity : eventEdges.keySet()) {
			List<ModelEdge> modelEdges = new ArrayList<ModelEdge>(sourceActivity.getOutputModelEdges());
			for (ModelEdge edge : modelEdges) {
				if (edge.getWeight() < 0.5) {
					Integer iof = valuesBelowHalf.indexOf(edge.getWeight());
					double correctedValue = 0.5 * (new Float(iof))/(new Float(valuesBelowHalf.size()));
					edge.setWeight(correctedValue);
				}
			}
		}
	}
	
	public void calculateObjectIntersectionDummy() {
		for (XOCObject o1 : this.xocObjects.values()) {
			this.objectsIntersections.put(o1, new ArrayList<XOCObject>());
		}
	}
	
	public void calculateObjectIntersection() {
		for (XOCObject o1 : this.xocObjects.values()) {
			this.objectsIntersections.put(o1, new ArrayList<XOCObject>());
			List<XOCEvent> e1 = new ArrayList<XOCEvent>(o1.getSubscribedEvents());
			if (e1.size() > 0) {
				for (XOCObject o2 : this.xocObjects.values()) {
					if (!o1.getObjectName().equals(o2.getObjectName())) {
						List<XOCEvent> e2 = new ArrayList<XOCEvent>(o2.getSubscribedEvents());
						if (e2.size() > 0) {
							e1 = new ArrayList<XOCEvent>(o1.getSubscribedEvents());
							e1.retainAll(e2);
							if (e1.size() > 0) {
								this.objectsIntersections.get(o1).add(o2);
							}
						}
					}
				}
			}
			
			if (this.objectsIntersections.get(o1).size() > 0) {
				System.out.println(o1.getObjectName());
				System.out.println("intersections: "+this.objectsIntersections.get(o1).size());
				System.out.println("");
			}
		}
	}
	
	public Set<String> getObjectsInvolvedInActivity(String activity, String wantedObjectClass) {
		Set<String> involvedObjectsIds = new HashSet<String>();
		if (this.xocActivities.containsKey(activity)) {
			Set<XOCObject> involvedObjects = this.xocActivities.get(activity).objectsInSubscribedEvents;
			for (XOCObject obj : involvedObjects) {
				if (wantedObjectClass == null || obj.getObjectClass().getClassLabel().equals(wantedObjectClass)) {
					involvedObjectsIds.add(obj.getObjectName());
				}
			}
		}
		return involvedObjectsIds;
	}
	
	public Set<String> getObjectsInvolvedInEdge(ModelEdge edge) {
		Set<String> involvedObjectsIds = new HashSet<String>();
		Set<XOCObject> involvedObjects = edge.getAssociatedObjects();
		
		for (XOCObject object : involvedObjects) {
			involvedObjectsIds.add(object.getObjectName());
		}
		
		return involvedObjectsIds;
	}
	
	public void filterXocOnActivity(String activity, String wantedObjectClass) {
		Set<String> objectsInvolved = getObjectsInvolvedInActivity(activity, wantedObjectClass);
		parseSimple(xocText, objectsInvolved);
	}
	
	public void filterXocOnSubgraph(Subgraph subgraph) {
		Set<String> objectsInvolved = subgraph.involvedObjectNames;
		parseSimple(xocText, objectsInvolved);
	}
	
	public void filterXocOnEdges(Set<ModelEdge> edges, String operation) {
		Set<XOCObject> objectsInvolved = null;
		
		for (ModelEdge edge : edges) {
			Set<XOCObject> edgeObjects = edge.getAssociatedObjects();
			if (objectsInvolved == null) {
				objectsInvolved = new HashSet<XOCObject>(edgeObjects);
			}
			else {
				if (operation == "AND") {
					objectsInvolved.retainAll(edgeObjects);
				}
				else if (operation == "XOR") {
					objectsInvolved.addAll(edgeObjects);
				}
			}
		}
		
		Set<String> objectsNamesInvolved = new HashSet<String>();
		
		for (XOCObject object : objectsInvolved) {
			objectsNamesInvolved.add(object.getObjectName());
		}
		
		parseSimple(xocText, objectsNamesInvolved);
	}
	
	public boolean checkEvent(String event, Set<String> listOfIdsToKeep) {
		if (listOfIdsToKeep == null) {
			return true;
		}
		List<String> ids = new ArrayList<String>(Arrays.asList(event.split("<string key=\"id\" value=\"")));
		if (ids.size() > 0) {
			ids.remove(0);
			int i = 0;
			while (i < ids.size()) {
				ids.set(i, ids.get(i).split("\"")[0]);
				i++;
			}
		}
		Set<String> idsSet = new HashSet<String>(ids);
		idsSet.retainAll(listOfIdsToKeep);
		if (idsSet.size() > 0) {
			return true;
		}
		return false;
	}
	
	public void parseSimple(String text, Set<String> listOfIdsToKeep) {
		if (!(listOfIdsToKeep == null)) {
			cloneIntoUnfiltered(false);
		}
		if (listOfIdsToKeep != null) {
			if (this.listOfIdsToKeep == null) {
				this.listOfIdsToKeep = listOfIdsToKeep;
			}
			else {
				this.listOfIdsToKeep.retainAll(listOfIdsToKeep);
			}
		}
		initializeModel();
		this.xocText = text;

		String[] events = text.split("<event>");
		int i = 1;
		while (i < events.length) {
			String event = events[i];
			if (checkEvent(event, listOfIdsToKeep)) {
				String eventId = event.split("<string key=\"id\" value=\"")[1].split("\"")[0];
				String eventActivity = event.split("<string key=\"activity\" value=\"")[1].split("\"")[0];
				if (this.wantedActivities == null || this.wantedActivities.contains(eventActivity)) {
					String eventTimestamp = event.split("<string key=\"timestamp\" value=\"")[1].split("\"")[0];
					SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss");
					Date parsedDate = null;
					try {
						parsedDate = dateFormat.parse(eventTimestamp);
					} catch (ParseException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					if (!xocActivities.containsKey(eventActivity)) {
						xocActivities.put(eventActivity, new XOCActivity(eventActivity));
					}
					XOCEvent xocEvent = new XOCEvent(eventId, parsedDate, xocActivities.get(eventActivity));
					orderedXocEvents.add(xocEvent);
					xocEvents.put(eventId, xocEvent);
					String[] objectReferences = event.split("<references>")[1].split("</references>")[0].split("<object>");
					int j = 1;
					while (j < objectReferences.length) {
						String objectId = objectReferences[j].split("<string key=\"id\" value=\"")[1].split("\"")[0];
						if (!xocObjects.containsKey(objectId)) {
							xocObjects.put(objectId, new XOCObject(objectId, xocObjectClasses));
						}
						xocEvents.get(eventId).addReferredObject(xocObjects.get(objectId));
						j++;
					}
					if (event.contains("<relations>") && event.contains("</relations>") && event.contains("<relation>")) {
						j = 1;
						String[] relations = event.split("<relations>")[1].split("<\relations>")[0].split("<relation>");
						while (j < relations.length) {
							String sourceId = relations[j].split("<string key=\"sourceobjectid\" value=\"")[1].split("\"")[0];
							String targetId = relations[j].split("<string key=\"targetobjectid\" value=\"")[1].split("\"")[0];
							if (!xocObjects.containsKey(sourceId)) {
								xocObjects.put(sourceId, new XOCObject(sourceId, xocObjectClasses));
							}
							if (!xocObjects.containsKey(targetId)) {
								xocObjects.put(targetId, new XOCObject(targetId, xocObjectClasses));
							}
							xocEvents.get(eventId).addReferredObject(xocObjects.get(sourceId));
							xocEvents.get(eventId).addReferredObject(xocObjects.get(targetId));
							XOCRelation xocRelation = new XOCRelation(xocEvents.get(eventId), xocObjects.get(sourceId), xocObjects.get(targetId));
							xocEvents.get(eventId).addRelation(xocRelation);
							j++;
						}
					}
					String[] objectModel = event.split("<model>")[1].split("<objects>")[1].split("<object>");
					j = 1;
					while (j < objectModel.length) {
						String objAttrValues = objectModel[j].split("</object>")[0];
						String objId = objAttrValues.split("<string key=\"id\" value=\"")[1].split("\"")[0];
						String objClass = objAttrValues.split("<string key=\"class\" value=\"")[1].split("\"")[0];
						XOCObjectVersion xocObjectVersion = new XOCObjectVersion(objId, xocObjectClasses);
						String[] attributes = objAttrValues.split("<string key=\"");
						int z = 1;
						while (z < attributes.length) {
							try {
								String attributeKey = attributes[z].split("\"")[0];
								String attributeValue = attributes[z].split("value=\"")[1].split("\"")[0];
								xocObjectVersion.addAttribute(attributeKey, attributeValue);
							}
							catch (Exception exx) {
								exx.printStackTrace();
							}
							z++;
						}
						xocEvents.get(eventId).addXocObjectVersion(xocObjectVersion);
						j++;
					}
				}
			}
			
			i++;
			
			if (i%10000 == 0) {
				System.out.println("read event "+i);
			}
		}
		
		System.out.println("finding event edges");
		findEventEdges();
		System.out.println("mining model");
		mineModel();
		System.out.println("checking edges importance");
		checkEdgesImportance();
		System.out.println("revising edges weight");
		reviseEdgesWeight();
		System.out.println("revising edges weight 2");
		reviseEdgesWeight2();
		System.out.println("calculating object intersection");
		if (this.xocObjects.size() > 70000) {
			this.calculateObjectIntersectionDummy();
		}
		else {
			this.calculateObjectIntersectionDummy();
			//calculateObjectIntersection();
		}
		
		for (String activityName : this.xocActivities.keySet()) {
			XOCActivity activity = this.xocActivities.get(activityName);
			activity.calculateRelationshipObjects();
			activity.calculateObjectsInSubscribedEvents();
			//activity.printRelationshipObjects();
		}
		
		System.out.println(orderedXocEvents.size());
	}
		
	public void parseSimple(InputStream stream, Set<String> listOfIdsToKeep) {
		StringWriter writer = new StringWriter();
		try {
			IOUtils.copy(stream, writer);
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		String text = writer.toString();
		
		parseSimple(text, listOfIdsToKeep);
	}
	
	public void parseFromSqliteDatabase(String filename) {
		String url = "jdbc:sqlite:"+filename.replace("\\", "/");
		
		Connection conn = null;
		try {
			conn = DriverManager.getConnection(url);
			PreparedStatement classStatement = conn.prepareStatement("SELECT id, name FROM class");
			ResultSet classResults = classStatement.executeQuery();
			
			Map<Integer, String> classIdToName = new HashMap<Integer, String>();
			while (classResults.next()) {
				int classId = classResults.getInt(1);
				String className = classResults.getString(2);
				classIdToName.put(classId, className);
				
				XOCObjectClass objectClass = new XOCObjectClass(className);
				xocObjectClasses.put(className, objectClass);
			}
			
			
			PreparedStatement objectStatement = conn.prepareStatement("SELECT id, class_id FROM object");
			ResultSet objectResults = objectStatement.executeQuery();
			while (objectResults.next()) {
				String objectId = ""+objectResults.getInt(1);
				String objectClassString = classIdToName.get(objectResults.getInt(2));
				XOCObjectClass objectClass = xocObjectClasses.get(objectClassString);
				XOCObject newObject = new XOCObject(objectId, xocObjectClasses);
				newObject.setObjectClass(objectClass);
				objectClass.subscribeObject(newObject);
				xocObjects.put(objectId, newObject);
			}
			
			PreparedStatement activityStatement = conn.prepareStatement("SELECT id, name FROM activity");
			ResultSet activityResults = activityStatement.executeQuery();
			Map<Integer, String> activityIdToName = new HashMap<Integer, String>();
			while (activityResults.next()) {
				int activityId = activityResults.getInt(1);
				String activityName = activityResults.getString(2);
				activityIdToName.put(activityId, activityName);
				
				XOCActivity activity = new XOCActivity(activityName);
				xocActivities.put(activityName, activity);
			}
						
			PreparedStatement activityInstance = conn.prepareStatement("SELECT id, activity_id FROM activity_instance");
			ResultSet activityInstanceResults = activityInstance.executeQuery();
			Map<Integer, Integer> activityInstanceIdMap = new HashMap<Integer, Integer>();
			while (activityInstanceResults.next()) {
				int activityInstanceId = activityInstanceResults.getInt(1);
				int activityActivityId = activityInstanceResults.getInt(2);
				activityInstanceIdMap.put(activityInstanceId, activityActivityId);
			}
			
			PreparedStatement eventStatement = conn.prepareStatement("SELECT id, activity_instance_id, timestamp FROM event");
			ResultSet eventResults = eventStatement.executeQuery();
			while (eventResults.next()) {
				String eventId = ""+eventResults.getInt(1);
				int activityId = eventResults.getInt(2);
				
				String activityName = activityIdToName.get(activityInstanceIdMap.get(activityId));
				XOCActivity activity = xocActivities.get(activityName);
				
				long timestamp = eventResults.getLong(3);
				
				Date newDate = new Date(timestamp);
				
				XOCEvent newEvent = new XOCEvent(eventId, newDate, activity);
				activity.addSubscribedEvent(newEvent);
				xocEvents.put(eventId, newEvent);
			}
			
			PreparedStatement objectVersionStatement = conn.prepareStatement("SELECT id, object_id FROM object_version");
			ResultSet objectVersionResults = objectVersionStatement.executeQuery();
			Map<Integer, Integer> objectVersionToObjectId = new HashMap<Integer, Integer>();
			while (objectVersionResults.next()) {
				int objectVersionId = objectVersionResults.getInt(1);
				int objectId = objectVersionResults.getInt(2);
				objectVersionToObjectId.put(objectVersionId, objectId);
			}
			
			//System.out.println(objectVersionToObjectId);
			
			PreparedStatement eventToObject = conn.prepareStatement("SELECT event_id, object_version_id FROM event_to_object_version");
			ResultSet eventToObjectResults = eventToObject.executeQuery();
			while (eventToObjectResults.next()) {
				String eventId = ""+eventToObjectResults.getInt(1);
				int objectVersionId = eventToObjectResults.getInt(2);
				String objectId = ""+objectVersionToObjectId.get(objectVersionId);
				
				XOCEvent event = xocEvents.get(eventId);
				XOCObject object = xocObjects.get(objectId);
				event.addReferredObject(object);
			}
			
			PreparedStatement relations = conn.prepareStatement("SELECT id, source_object_version_id, target_object_version_id FROM relation");
			ResultSet relationsResults = relations.executeQuery();
			while (relationsResults.next()) {
				int relationId = relationsResults.getInt(1);
				int sourceObjectVersionId = relationsResults.getInt(2);
				int targetObjectVersionId = relationsResults.getInt(3);
				String sourceObjectId = ""+objectVersionToObjectId.get(sourceObjectVersionId);
				String targetObjectId = ""+objectVersionToObjectId.get(targetObjectVersionId);
				
				XOCObject sourceObject = xocObjects.get(sourceObjectId);
				XOCObject targetObject = xocObjects.get(targetObjectId);
				
				for (XOCEvent ev : targetObject.getSubscribedEvents()) {
					ev.addReferredObject(sourceObject);
					//xocObjects.get(sourceObjectId).addSubscribedEvent(ev);
				}
				for (XOCEvent ev : sourceObject.getSubscribedEvents()) {
					ev.addReferredObject(targetObject);
					//xocObjects.get(targetObjectId).addSubscribedEvent(ev);
				}
				
				XOCRelation newRelation = new XOCRelation(null, sourceObject, targetObject);
				xocRelations.add(newRelation);
				System.out.println(xocRelations.size());
			}
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        if (conn != null) {
            try {
				conn.close();
			} catch (SQLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
        }
        
        /*for (XOCEvent event : xocEvents.values()) {
        	event.printResume();
        }*/
        
        System.out.println("finding model edges");
		findEventEdges();
		System.out.println("mining model");
		mineModel();
		System.out.println("checking edges importance");
		checkEdgesImportance();
		System.out.println("revising edges weight 1");
		reviseEdgesWeight();
		System.out.println("revising edges weight 2");
		reviseEdgesWeight2();
	}
	
	public List<XOCActivity> getActivities() {
		return new ArrayList<XOCActivity>(this.xocActivities.values());
	}
	
	public void removeDuplicates(List<Subgraph> subgraphs) {
		Set<Set<ModelEdge>> alreadySeenSubgraphs = new HashSet<Set<ModelEdge>>();
		int i = 0;
		while (i < subgraphs.size()) {
			Subgraph subgraph = subgraphs.get(i);
			Set<ModelEdge> involvedEdges = new HashSet<ModelEdge>(subgraph.involvedEdges);
			if (alreadySeenSubgraphs.contains(involvedEdges)) {
				subgraphs.remove(i);
				continue;
			}
			alreadySeenSubgraphs.add(involvedEdges);
			i++;
		}
	}
	
	public List<Subgraph> discoverMostInterestingSubgraph(String activityName) {
		List<Subgraph> subgraphs = new ArrayList<Subgraph>();
		XOCActivity activity = this.xocActivities.get(activityName);
		
		for (ModelEdge edge : activity.getOutputModelEdges()) {
			Subgraph subgraph = new Subgraph(this);
			subgraph.addEdge(edge);
			subgraph.calculateInvolvedEdgesAndScore();
			subgraph.setGraphBasicDirection("forward");
			subgraphs.add(subgraph);
		}
		for (ModelEdge edge : activity.getInputModelEdges()) {
			Subgraph subgraph = new Subgraph(this);
			subgraph.addEdge(edge);
			subgraph.calculateInvolvedEdgesAndScore();
			subgraph.setGraphBasicDirection("backward");
			subgraphs.add(subgraph);
		}
		
		int z = 0;
		int noOfRequiredIterationsAND = 2;
		while (z < noOfRequiredIterationsAND) {
			int subgraphsActualSize = subgraphs.size();
			int j = 0;
			while (j < subgraphsActualSize) {
				Subgraph subgraph = subgraphs.get(j);
				if (subgraph.graphOperator == "AND") {
					Set<ModelEdge> expansionEdgesForwardForward = subgraph.getPossibleExpansionEdgesAND("forwardForward");
					Set<ModelEdge> expansionEdgesForwardBackward = subgraph.getPossibleExpansionEdgesAND("forwardBackward");
					Set<ModelEdge> expansionEdgesBackwardForward = subgraph.getPossibleExpansionEdgesAND("backwardForward");
					Set<ModelEdge> expansionEdgesBackwardBackward = subgraph.getPossibleExpansionEdgesAND("backwardBackward");
					
					Set<ModelEdge> expansionEdges = new HashSet<ModelEdge>(expansionEdgesForwardForward);
					expansionEdges.addAll(expansionEdgesForwardBackward);
					expansionEdges.addAll(expansionEdgesBackwardForward);
					expansionEdges.addAll(expansionEdgesBackwardBackward);
									
					for (ModelEdge ee : expansionEdges) {
						Subgraph newSubgraph = new Subgraph(subgraph);
						newSubgraph.addEdge(ee);
						newSubgraph.calculateInvolvedEdgesAndScore();
						subgraphs.add(newSubgraph);
					}
				}
				j++;
			}
			
			removeDuplicates(subgraphs);
			Collections.sort(subgraphs, new SubgraphComparator());
			
			z++;
		}
		
		z = 0;
		int noOfRequiredIterationsOR = 1;
		while (z < noOfRequiredIterationsOR) {
			int subgraphsActualSize = subgraphs.size();
			int j = 0;
			while (j < subgraphsActualSize) {
				Subgraph subgraph = subgraphs.get(j);
				Set<ModelEdge> expansionEdges = subgraph.getPossibleExpansionEdgesOR();
				
				for (ModelEdge ee : expansionEdges) {
					Subgraph newSubgraph = new Subgraph(subgraph);
					newSubgraph.addEdge(ee);
					newSubgraph.setGraphOperator("XOR");
					newSubgraph.calculateInvolvedEdgesAndScore();
					subgraphs.add(newSubgraph);
				}
				j++;
			}
			removeDuplicates(subgraphs);
			Collections.sort(subgraphs, new SubgraphComparator());
			
			z++;
		}
				
		/*for (Subgraph subgraph : subgraphs) {
			subgraph.printResume();
		}*/
		
		return subgraphs;
	}
	
	public void printResume() {
		System.out.println("ParsedXOC events = "+this.xocEvents.size()+" objects = "+this.xocObjects.size());
	}
	
	public static void main(String[] args) {
		ParsedXOC parser = new ParsedXOC();
		try {
			parser.parseSimple(new FileInputStream("C:\\log.xoc"), null);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	
	
}
